package sun.misc;

import java.io.OutputStream;
import java.lang.reflect.Array;
import sun.security.action.GetBooleanAction;
import java.util.ListIterator;
import java.util.LinkedList;
import java.util.Iterator;
import java.io.DataOutputStream;
import java.io.ByteArrayOutputStream;
import java.util.HashMap;
import java.util.ArrayList;
import java.security.AccessController;
import java.nio.file.Path;
import java.io.IOException;
import java.nio.file.OpenOption;
import java.nio.file.Files;
import java.nio.file.attribute.FileAttribute;
import java.io.File;
import java.nio.file.Paths;
import java.security.PrivilegedAction;
import java.util.Map;
import java.util.List;
import java.lang.reflect.Method;

public class ProxyGenerator
{
    private static final int CLASSFILE_MAJOR_VERSION = 49;
    private static final int CLASSFILE_MINOR_VERSION = 0;
    private static final int CONSTANT_UTF8 = 1;
    private static final int CONSTANT_UNICODE = 2;
    private static final int CONSTANT_INTEGER = 3;
    private static final int CONSTANT_FLOAT = 4;
    private static final int CONSTANT_LONG = 5;
    private static final int CONSTANT_DOUBLE = 6;
    private static final int CONSTANT_CLASS = 7;
    private static final int CONSTANT_STRING = 8;
    private static final int CONSTANT_FIELD = 9;
    private static final int CONSTANT_METHOD = 10;
    private static final int CONSTANT_INTERFACEMETHOD = 11;
    private static final int CONSTANT_NAMEANDTYPE = 12;
    private static final int ACC_PUBLIC = 1;
    private static final int ACC_PRIVATE = 2;
    private static final int ACC_STATIC = 8;
    private static final int ACC_FINAL = 16;
    private static final int ACC_SUPER = 32;
    private static final int opc_aconst_null = 1;
    private static final int opc_iconst_0 = 3;
    private static final int opc_bipush = 16;
    private static final int opc_sipush = 17;
    private static final int opc_ldc = 18;
    private static final int opc_ldc_w = 19;
    private static final int opc_iload = 21;
    private static final int opc_lload = 22;
    private static final int opc_fload = 23;
    private static final int opc_dload = 24;
    private static final int opc_aload = 25;
    private static final int opc_iload_0 = 26;
    private static final int opc_lload_0 = 30;
    private static final int opc_fload_0 = 34;
    private static final int opc_dload_0 = 38;
    private static final int opc_aload_0 = 42;
    private static final int opc_astore = 58;
    private static final int opc_astore_0 = 75;
    private static final int opc_aastore = 83;
    private static final int opc_pop = 87;
    private static final int opc_dup = 89;
    private static final int opc_ireturn = 172;
    private static final int opc_lreturn = 173;
    private static final int opc_freturn = 174;
    private static final int opc_dreturn = 175;
    private static final int opc_areturn = 176;
    private static final int opc_return = 177;
    private static final int opc_getstatic = 178;
    private static final int opc_putstatic = 179;
    private static final int opc_getfield = 180;
    private static final int opc_invokevirtual = 182;
    private static final int opc_invokespecial = 183;
    private static final int opc_invokestatic = 184;
    private static final int opc_invokeinterface = 185;
    private static final int opc_new = 187;
    private static final int opc_anewarray = 189;
    private static final int opc_athrow = 191;
    private static final int opc_checkcast = 192;
    private static final int opc_wide = 196;
    private static final String superclassName = "java/lang/reflect/Proxy";
    private static final String handlerFieldName = "h";
    private static final boolean saveGeneratedFiles;
    private static Method hashCodeMethod;
    private static Method equalsMethod;
    private static Method toStringMethod;
    private String className;
    private Class<?>[] interfaces;
    private int accessFlags;
    private ProxyGenerator.ConstantPool cp;
    private List<ProxyGenerator.FieldInfo> fields;
    private List<ProxyGenerator.MethodInfo> methods;
    private Map<String, List<ProxyGenerator.ProxyMethod>> proxyMethods;
    private int proxyMethodCount;
    static final /* synthetic */ boolean $assertionsDisabled;
    
    public static byte[] generateProxyClass(String s, Class<?>[] array) {
        return generateProxyClass(s, array, 49);
    }
    
    public static byte[] generateProxyClass(final String s, Class<?>[] array, int n) {
        final byte[] generateClassFile = new ProxyGenerator(s, array, n).generateClassFile();
        if (ProxyGenerator.saveGeneratedFiles) {
            AccessController.doPrivileged(new PrivilegedAction<Void>() {
                @Override
                public Void run() {
                    try {
                        int lastIndex = s.lastIndexOf(46);
                        Path path;
                        if (lastIndex > 0) {
                            Path value = Paths.get(s.substring(0, lastIndex).replace('.', File.separatorChar), new String[0]);
                            Files.createDirectories(value, new FileAttribute[0]);
                            path = value.resolve(s.substring(lastIndex + 1, s.length()) + ".class");
                        }
                        else {
                            path = Paths.get(s + ".class", new String[0]);
                        }
                        Files.write(path, generateClassFile, new OpenOption[0]);
                        return null;
                    }
                    catch (IOException obj) {
                        throw new InternalError("I/O exception saving generated file: " + obj);
                    }
                }
            });
            return generateClassFile;
        }
        return generateClassFile;
    }
    
    private ProxyGenerator(String className, Class<?>[] interfaces, int accessFlags) {
        this.cp = new ProxyGenerator.ConstantPool();
        this.fields = new ArrayList();
        this.methods = new ArrayList();
        this.proxyMethods = new HashMap();
        this.proxyMethodCount = 0;
        this.className = className;
        this.interfaces = interfaces;
        this.accessFlags = accessFlags;
    }
    
    private byte[] generateClassFile() {
        this.addProxyMethod(ProxyGenerator.hashCodeMethod, Object.class);
        this.addProxyMethod(ProxyGenerator.equalsMethod, Object.class);
        this.addProxyMethod(ProxyGenerator.toStringMethod, Object.class);
        for (Class clazz : this.interfaces) {
            Method[] methods = clazz.getMethods();
            for (int length2 = methods.length, j = 0; j < length2; ++j) {
                this.addProxyMethod(methods[j], clazz);
            }
        }
        Iterator<List<ProxyGenerator.ProxyMethod>> iterator = this.proxyMethods.values().iterator();
        while (iterator.hasNext()) {
            checkReturnTypes((List)iterator.next());
        }
        try {
            this.methods.add(this.generateConstructor());
            Iterator<List<ProxyGenerator.ProxyMethod>> iterator2 = this.proxyMethods.values().iterator();
            while (iterator2.hasNext()) {
                for (ProxyGenerator.ProxyMethod proxyMethod : (List)iterator2.next()) {
                    this.fields.add(new ProxyGenerator.FieldInfo(proxyMethod.methodFieldName, "Ljava/lang/reflect/Method;", 10));
                    this.methods.add(ProxyGenerator.ProxyMethod.access$100(proxyMethod));
                }
            }
            this.methods.add(this.generateStaticInitializer());
        }
        catch (IOException cause) {
            throw new InternalError("unexpected I/O Exception", cause);
        }
        if (this.methods.size() > 65535) {
            throw new IllegalArgumentException("method limit exceeded");
        }
        if (this.fields.size() > 65535) {
            throw new IllegalArgumentException("field limit exceeded");
        }
        this.cp.getClass(dotToSlash(this.className));
        this.cp.getClass("java/lang/reflect/Proxy");
        Class[] interfaces2 = this.interfaces;
        for (int length3 = interfaces2.length, k = 0; k < length3; ++k) {
            this.cp.getClass(dotToSlash(interfaces2[k].getName()));
        }
        this.cp.setReadOnly();
        ByteArrayOutputStream out = new ByteArrayOutputStream();
        DataOutputStream o = new DataOutputStream(out);
        try {
            o.writeInt(-889275714);
            o.writeShort(0);
            o.writeShort(49);
            this.cp.write(o);
            o.writeShort(this.accessFlags);
            o.writeShort((int)this.cp.getClass(dotToSlash(this.className)));
            o.writeShort((int)this.cp.getClass("java/lang/reflect/Proxy"));
            o.writeShort(this.interfaces.length);
            Class[] interfaces3 = this.interfaces;
            for (int length4 = interfaces3.length, l = 0; l < length4; ++l) {
                o.writeShort((int)this.cp.getClass(dotToSlash(interfaces3[l].getName())));
            }
            o.writeShort(this.fields.size());
            Iterator<ProxyGenerator.FieldInfo> iterator4 = this.fields.iterator();
            while (iterator4.hasNext()) {
                ((ProxyGenerator.FieldInfo)iterator4.next()).write(o);
            }
            o.writeShort(this.methods.size());
            Iterator<ProxyGenerator.MethodInfo> iterator5 = this.methods.iterator();
            while (iterator5.hasNext()) {
                ((ProxyGenerator.MethodInfo)iterator5.next()).write(o);
            }
            o.writeShort(0);
        }
        catch (IOException cause2) {
            throw new InternalError("unexpected I/O Exception", cause2);
        }
        return out.toByteArray();
    }
    
    private void addProxyMethod(Method method, Class<?> clazz) {
        String name = method.getName();
        Class[] parameterTypes = method.getParameterTypes();
        Class<Object> returnType = method.getReturnType();
        Class[] exceptionTypes = method.getExceptionTypes();
        String string = name + getParameterDescriptors(parameterTypes);
        List<ProxyGenerator.ProxyMethod> list = (List)this.proxyMethods.get(string);
        if (list != null) {
            for (ProxyGenerator.ProxyMethod proxyMethod : list) {
                if (returnType == proxyMethod.returnType) {
                    ArrayList<Class<?>> list2 = new ArrayList();
                    collectCompatibleTypes(exceptionTypes, proxyMethod.exceptionTypes, list2);
                    collectCompatibleTypes(proxyMethod.exceptionTypes, exceptionTypes, list2);
                    proxyMethod.exceptionTypes = new Class[list2.size()];
                    proxyMethod.exceptionTypes = (Class[])list2.toArray(proxyMethod.exceptionTypes);
                    return;
                }
                continue;
            }
        }
        else {
            list = new ArrayList(3);
            this.proxyMethods.put(string, list);
        }
        list.add(new ProxyGenerator.ProxyMethod(name, parameterTypes, returnType, exceptionTypes, clazz));
    }
    
    private static void checkReturnTypes(List<ProxyGenerator.ProxyMethod> list) {
        if (list.size() < 2) {
            return;
        }
        LinkedList<Class> obj = new LinkedList();
    Label_0026:
        for (ProxyGenerator.ProxyMethod proxyMethod : list) {
            Class<Object> returnType = proxyMethod.returnType;
            if (returnType.isPrimitive()) {
                throw new IllegalArgumentException("methods with same signature " + getFriendlyMethodSignature(proxyMethod.methodName, proxyMethod.parameterTypes) + " but incompatible return types: " + returnType.getName() + " and others");
            }
            int n = 0;
            ListIterator<Class> listIterator = obj.listIterator();
            while (listIterator.hasNext()) {
                Class<Object> clazz = (Class)listIterator.next();
                if (returnType.isAssignableFrom(clazz)) {
                    if (ProxyGenerator.$assertionsDisabled || n == 0) {
                        continue Label_0026;
                    }
                    throw new AssertionError();
                }
                else if (clazz.isAssignableFrom(returnType)) {
                    if (n == 0) {
                        listIterator.set(returnType);
                        n = 1;
                    }
                    else {
                        listIterator.remove();
                    }
                }
                else {
                    continue;
                }
            }
            if (n == 0) {
                obj.add(returnType);
            }
            else {
                continue;
            }
        }
        if (obj.size() > 1) {
            ProxyGenerator.ProxyMethod proxyMethod2 = (ProxyGenerator.ProxyMethod)list.get(0);
            throw new IllegalArgumentException("methods with same signature " + getFriendlyMethodSignature(proxyMethod2.methodName, proxyMethod2.parameterTypes) + " but incompatible return types: " + obj);
        }
        return;
    }
    
    private ProxyGenerator.MethodInfo generateConstructor() throws IOException {
        ProxyGenerator.MethodInfo methodInfo = new ProxyGenerator.MethodInfo("<init>", "(Ljava/lang/reflect/InvocationHandler;)V", 1);
        DataOutputStream dataOutputStream = new DataOutputStream(methodInfo.code);
        this.code_aload(0, dataOutputStream);
        this.code_aload(1, dataOutputStream);
        dataOutputStream.writeByte(183);
        dataOutputStream.writeShort((int)this.cp.getMethodRef("java/lang/reflect/Proxy", "<init>", "(Ljava/lang/reflect/InvocationHandler;)V"));
        dataOutputStream.writeByte(177);
        methodInfo.maxStack = 10;
        methodInfo.maxLocals = 2;
        methodInfo.declaredExceptions = new short[0];
        return methodInfo;
    }
    
    private ProxyGenerator.MethodInfo generateStaticInitializer() throws IOException {
        ProxyGenerator.MethodInfo methodInfo = new ProxyGenerator.MethodInfo("<clinit>", "()V", 8);
        int n = 1;
        int n2 = 0;
        DataOutputStream dataOutputStream = new DataOutputStream(methodInfo.code);
        Iterator<List<ProxyGenerator.ProxyMethod>> iterator = this.proxyMethods.values().iterator();
        while (iterator.hasNext()) {
            Iterator<Object> iterator2 = ((List)iterator.next()).iterator();
            while (iterator2.hasNext()) {
                ProxyGenerator.ProxyMethod.access$1900((ProxyGenerator.ProxyMethod)iterator2.next(), dataOutputStream);
            }
        }
        dataOutputStream.writeByte(177);
        int n3;
        methodInfo.exceptionTable.add(new ProxyGenerator.ExceptionTableEntry((short)n2, (short)n3, (short)(n3 = (short)methodInfo.code.size()), this.cp.getClass("java/lang/NoSuchMethodException")));
        this.code_astore(n, dataOutputStream);
        dataOutputStream.writeByte(187);
        dataOutputStream.writeShort((int)this.cp.getClass("java/lang/NoSuchMethodError"));
        dataOutputStream.writeByte(89);
        this.code_aload(n, dataOutputStream);
        dataOutputStream.writeByte(182);
        dataOutputStream.writeShort((int)this.cp.getMethodRef("java/lang/Throwable", "getMessage", "()Ljava/lang/String;"));
        dataOutputStream.writeByte(183);
        dataOutputStream.writeShort((int)this.cp.getMethodRef("java/lang/NoSuchMethodError", "<init>", "(Ljava/lang/String;)V"));
        dataOutputStream.writeByte(191);
        methodInfo.exceptionTable.add(new ProxyGenerator.ExceptionTableEntry((short)n2, (short)n3, (short)methodInfo.code.size(), this.cp.getClass("java/lang/ClassNotFoundException")));
        this.code_astore(n, dataOutputStream);
        dataOutputStream.writeByte(187);
        dataOutputStream.writeShort((int)this.cp.getClass("java/lang/NoClassDefFoundError"));
        dataOutputStream.writeByte(89);
        this.code_aload(n, dataOutputStream);
        dataOutputStream.writeByte(182);
        dataOutputStream.writeShort((int)this.cp.getMethodRef("java/lang/Throwable", "getMessage", "()Ljava/lang/String;"));
        dataOutputStream.writeByte(183);
        dataOutputStream.writeShort((int)this.cp.getMethodRef("java/lang/NoClassDefFoundError", "<init>", "(Ljava/lang/String;)V"));
        dataOutputStream.writeByte(191);
        if (methodInfo.code.size() > 65535) {
            throw new IllegalArgumentException("code size limit exceeded");
        }
        methodInfo.maxStack = 10;
        methodInfo.maxLocals = (short)(n + 1);
        methodInfo.declaredExceptions = new short[0];
        return methodInfo;
    }
    
    private void code_iload(int n, DataOutputStream dataOutputStream) throws IOException {
        this.codeLocalLoadStore(n, 21, 26, dataOutputStream);
    }
    
    private void code_lload(int n, DataOutputStream dataOutputStream) throws IOException {
        this.codeLocalLoadStore(n, 22, 30, dataOutputStream);
    }
    
    private void code_fload(int n, DataOutputStream dataOutputStream) throws IOException {
        this.codeLocalLoadStore(n, 23, 34, dataOutputStream);
    }
    
    private void code_dload(int n, DataOutputStream dataOutputStream) throws IOException {
        this.codeLocalLoadStore(n, 24, 38, dataOutputStream);
    }
    
    private void code_aload(int n, DataOutputStream dataOutputStream) throws IOException {
        this.codeLocalLoadStore(n, 25, 42, dataOutputStream);
    }
    
    private void code_astore(int n, DataOutputStream dataOutputStream) throws IOException {
        this.codeLocalLoadStore(n, 58, 75, dataOutputStream);
    }
    
    private void codeLocalLoadStore(int n, int n2, int n3, DataOutputStream dataOutputStream) throws IOException {
        assert n >= 0 && n <= 65535;
        if (n <= 3) {
            dataOutputStream.writeByte(n3 + n);
        }
        else if (n <= 255) {
            dataOutputStream.writeByte(n2);
            dataOutputStream.writeByte(n & 0xFF);
        }
        else {
            dataOutputStream.writeByte(196);
            dataOutputStream.writeByte(n2);
            dataOutputStream.writeShort(n & 0xFFFF);
        }
        return;
    }
    
    private void code_ldc(int n, DataOutputStream dataOutputStream) throws IOException {
        assert n >= 0 && n <= 65535;
        if (n <= 255) {
            dataOutputStream.writeByte(18);
            dataOutputStream.writeByte(n & 0xFF);
        }
        else {
            dataOutputStream.writeByte(19);
            dataOutputStream.writeShort(n & 0xFFFF);
        }
        return;
    }
    
    private void code_ipush(int n, DataOutputStream dataOutputStream) throws IOException {
        if (n >= -1 && n <= 5) {
            dataOutputStream.writeByte(3 + n);
        }
        else if (n >= -128 && n <= 127) {
            dataOutputStream.writeByte(16);
            dataOutputStream.writeByte(n & 0xFF);
        }
        else if (n >= -32768 && n <= 32767) {
            dataOutputStream.writeByte(17);
            dataOutputStream.writeShort(n & 0xFFFF);
        }
        else {
            throw new AssertionError();
        }
    }
    
    private void codeClassForName(Class<?> clazz, DataOutputStream dataOutputStream) throws IOException {
        this.code_ldc((int)this.cp.getString(clazz.getName()), dataOutputStream);
        dataOutputStream.writeByte(184);
        dataOutputStream.writeShort((int)this.cp.getMethodRef("java/lang/Class", "forName", "(Ljava/lang/String;)Ljava/lang/Class;"));
    }
    
    private static String dotToSlash(String s) {
        return s.replace('.', '/');
    }
    
    private static String getMethodDescriptor(Class<?>[] array, Class<?> clazz) {
        return getParameterDescriptors(array) + ((clazz == Void.TYPE) ? "V" : getFieldType(clazz));
    }
    
    private static String getParameterDescriptors(Class<?>[] array) {
        StringBuilder sb = new StringBuilder("(");
        for (int i = 0; i < array.length; ++i) {
            sb.append(getFieldType(array[i]));
        }
        sb.append(')');
        return sb.toString();
    }
    
    private static String getFieldType(Class<?> clazz) {
        if (clazz.isPrimitive()) {
            return ProxyGenerator.PrimitiveTypeInfo.get(clazz).baseTypeString;
        }
        if (clazz.isArray()) {
            return clazz.getName().replace('.', '/');
        }
        return "L" + dotToSlash(clazz.getName()) + ";";
    }
    
    private static String getFriendlyMethodSignature(String str, Class<?>[] array) {
        StringBuilder sb = new StringBuilder(str);
        sb.append('(');
        for (int i = 0; i < array.length; ++i) {
            if (i > 0) {
                sb.append(',');
            }
            Class<?> componentType;
            int n;
            for (componentType = array[i], n = 0; componentType.isArray(); componentType = componentType.getComponentType(), ++n) {}
            sb.append(componentType.getName());
            while (n-- > 0) {
                sb.append("[]");
            }
        }
        sb.append(')');
        return sb.toString();
    }
    
    private static int getWordsPerType(Class<?> clazz) {
        if (clazz == Long.TYPE || clazz == Double.TYPE) {
            return 2;
        }
        return 1;
    }
    
    private static void collectCompatibleTypes(Class<?>[] array, Class<?>[] array2, List<Class<?>> list) {
        for (Class<?> clazz : array) {
            if (!list.contains(clazz)) {
                int length2 = array2.length;
                int j = 0;
                while (j < length2) {
                    if (array2[j].isAssignableFrom(clazz)) {
                        list.add(clazz);
                        break;
                    }
                    ++j;
                }
            }
        }
    }
    
    private static List<Class<?>> computeUniqueCatchList(Class<?>[] array) {
        ArrayList<Class<?>> list = new ArrayList();
        list.add(Error.class);
        list.add(RuntimeException.class);
        int length = array.length;
        int i = 0;
        while (i < length) {
            Class<?> clazz = array[i];
            if (clazz.isAssignableFrom(Throwable.class)) {
                list.clear();
                break;
            }
            Label_0155: {
                if (Throwable.class.isAssignableFrom(clazz)) {
                    int j = 0;
                    while (j < list.size()) {
                        Class<Object> clazz2 = (Class)list.get(j);
                        if (clazz2.isAssignableFrom(clazz)) {
                            break Label_0155;
                        }
                        if (clazz.isAssignableFrom(clazz2)) {
                            list.remove(j);
                        }
                        else {
                            ++j;
                        }
                    }
                    list.add(clazz);
                }
            }
            ++i;
        }
        return list;
    }
    
    static String access$500(Class[] array, Class clazz) {
        return getMethodDescriptor(array, clazz);
    }
    
    static int access$600(Class clazz) {
        return getWordsPerType(clazz);
    }
    
    static void access$700(ProxyGenerator proxyGenerator, int n, DataOutputStream dataOutputStream) throws IOException {
        proxyGenerator.code_aload(n, dataOutputStream);
    }
    
    static String access$900(String s) {
        return dotToSlash(s);
    }
    
    static void access$1000(ProxyGenerator proxyGenerator, int n, DataOutputStream dataOutputStream) throws IOException {
        proxyGenerator.code_ipush(n, dataOutputStream);
    }
    
    static List access$1100(Class[] array) {
        return computeUniqueCatchList(array);
    }
    
    static void access$1200(ProxyGenerator proxyGenerator, int n, DataOutputStream dataOutputStream) throws IOException {
        proxyGenerator.code_astore(n, dataOutputStream);
    }
    
    static void access$1300(ProxyGenerator proxyGenerator, int n, DataOutputStream dataOutputStream) throws IOException {
        proxyGenerator.code_iload(n, dataOutputStream);
    }
    
    static void access$1400(ProxyGenerator proxyGenerator, int n, DataOutputStream dataOutputStream) throws IOException {
        proxyGenerator.code_lload(n, dataOutputStream);
    }
    
    static void access$1500(ProxyGenerator proxyGenerator, int n, DataOutputStream dataOutputStream) throws IOException {
        proxyGenerator.code_fload(n, dataOutputStream);
    }
    
    static void access$1600(ProxyGenerator proxyGenerator, int n, DataOutputStream dataOutputStream) throws IOException {
        proxyGenerator.code_dload(n, dataOutputStream);
    }
    
    static void access$1700(ProxyGenerator proxyGenerator, Class clazz, DataOutputStream dataOutputStream) throws IOException {
        proxyGenerator.codeClassForName(clazz, dataOutputStream);
    }
    
    static void access$1800(ProxyGenerator proxyGenerator, int n, DataOutputStream dataOutputStream) throws IOException {
        proxyGenerator.code_ldc(n, dataOutputStream);
    }
    
    static {
        saveGeneratedFiles = (boolean)AccessController.doPrivileged(new GetBooleanAction("sun.misc.ProxyGenerator.saveGeneratedFiles"));
        try {
            ProxyGenerator.hashCodeMethod = Object.class.getMethod("hashCode", new Class[0]);
            ProxyGenerator.equalsMethod = Object.class.getMethod("equals", new Class[] { Object.class });
            ProxyGenerator.toStringMethod = Object.class.getMethod("toString", new Class[0]);
        }
        catch (NoSuchMethodException ex) {
            throw new NoSuchMethodError(ex.getMessage());
        }
    }
    
    private class FieldInfo
    {
        public int accessFlags;
        public String name;
        public String descriptor;
        
        public FieldInfo(String name, String descriptor, int accessFlags) {
            this.name = name;
            this.descriptor = descriptor;
            this.accessFlags = accessFlags;
            ProxyGenerator.this.cp.getUtf8(name);
            ProxyGenerator.this.cp.getUtf8(descriptor);
        }
        
        public void write(DataOutputStream dataOutputStream) throws IOException {
            dataOutputStream.writeShort(this.accessFlags);
            dataOutputStream.writeShort((int)ProxyGenerator.this.cp.getUtf8(this.name));
            dataOutputStream.writeShort((int)ProxyGenerator.this.cp.getUtf8(this.descriptor));
            dataOutputStream.writeShort(0);
        }
    }
    
    private static class ExceptionTableEntry
    {
        public short startPc;
        public short endPc;
        public short handlerPc;
        public short catchType;
        
        public ExceptionTableEntry(short startPc, short endPc, short handlerPc, short catchType) {
            this.startPc = startPc;
            this.endPc = endPc;
            this.handlerPc = handlerPc;
            this.catchType = catchType;
        }
    }
    
    private class MethodInfo
    {
        public int accessFlags;
        public String name;
        public String descriptor;
        public short maxStack;
        public short maxLocals;
        public ByteArrayOutputStream code;
        public List<ProxyGenerator.ExceptionTableEntry> exceptionTable;
        public short[] declaredExceptions;
        
        public MethodInfo(String name, String descriptor, int accessFlags) {
            this.code = new ByteArrayOutputStream();
            this.exceptionTable = new ArrayList();
            this.name = name;
            this.descriptor = descriptor;
            this.accessFlags = accessFlags;
            ProxyGenerator.this.cp.getUtf8(name);
            ProxyGenerator.this.cp.getUtf8(descriptor);
            ProxyGenerator.this.cp.getUtf8("Code");
            ProxyGenerator.this.cp.getUtf8("Exceptions");
        }
        
        public void write(DataOutputStream out) throws IOException {
            out.writeShort(this.accessFlags);
            out.writeShort((int)ProxyGenerator.this.cp.getUtf8(this.name));
            out.writeShort((int)ProxyGenerator.this.cp.getUtf8(this.descriptor));
            out.writeShort(2);
            out.writeShort((int)ProxyGenerator.this.cp.getUtf8("Code"));
            out.writeInt(12 + this.code.size() + 8 * this.exceptionTable.size());
            out.writeShort((int)this.maxStack);
            out.writeShort((int)this.maxLocals);
            out.writeInt(this.code.size());
            this.code.writeTo(out);
            out.writeShort(this.exceptionTable.size());
            for (ProxyGenerator.ExceptionTableEntry exceptionTableEntry : this.exceptionTable) {
                out.writeShort((int)exceptionTableEntry.startPc);
                out.writeShort((int)exceptionTableEntry.endPc);
                out.writeShort((int)exceptionTableEntry.handlerPc);
                out.writeShort((int)exceptionTableEntry.catchType);
            }
            out.writeShort(0);
            out.writeShort((int)ProxyGenerator.this.cp.getUtf8("Exceptions"));
            out.writeInt(2 + 2 * this.declaredExceptions.length);
            out.writeShort(this.declaredExceptions.length);
            short[] declaredExceptions = this.declaredExceptions;
            for (int length = declaredExceptions.length, i = 0; i < length; ++i) {
                out.writeShort((int)declaredExceptions[i]);
            }
        }
    }
    
    private class ProxyMethod
    {
        public String methodName;
        public Class<?>[] parameterTypes;
        public Class<?> returnType;
        public Class<?>[] exceptionTypes;
        public Class<?> fromClass;
        public String methodFieldName;
        
        private ProxyMethod(String methodName, Class<?>[] parameterTypes, Class<?> returnType, Class<?>[] exceptionTypes, Class<?> fromClass) {
            this.methodName = methodName;
            this.parameterTypes = parameterTypes;
            this.returnType = returnType;
            this.exceptionTypes = exceptionTypes;
            this.fromClass = fromClass;
            this.methodFieldName = "m" + ProxyGenerator.this.proxyMethodCount++;
        }
        
        private ProxyGenerator.MethodInfo generateMethod() throws IOException {
            ProxyGenerator.MethodInfo methodInfo = new ProxyGenerator.MethodInfo(this.methodName, ProxyGenerator.access$500(this.parameterTypes, this.returnType), 17);
            int[] array = new int[this.parameterTypes.length];
            int n = 1;
            for (int i = 0; i < array.length; ++i) {
                array[i] = n;
                n += ProxyGenerator.access$600(this.parameterTypes[i]);
            }
            int n2 = 0;
            DataOutputStream o = new DataOutputStream(methodInfo.code);
            ProxyGenerator.access$700(ProxyGenerator.this, 0, o);
            o.writeByte(180);
            o.writeShort((int)ProxyGenerator.this.cp.getFieldRef("java/lang/reflect/Proxy", "h", "Ljava/lang/reflect/InvocationHandler;"));
            ProxyGenerator.access$700(ProxyGenerator.this, 0, o);
            o.writeByte(178);
            o.writeShort((int)ProxyGenerator.this.cp.getFieldRef(ProxyGenerator.access$900(ProxyGenerator.this.className), this.methodFieldName, "Ljava/lang/reflect/Method;"));
            if (this.parameterTypes.length > 0) {
                ProxyGenerator.access$1000(ProxyGenerator.this, this.parameterTypes.length, o);
                o.writeByte(189);
                o.writeShort((int)ProxyGenerator.this.cp.getClass("java/lang/Object"));
                for (int j = 0; j < this.parameterTypes.length; ++j) {
                    o.writeByte(89);
                    ProxyGenerator.access$1000(ProxyGenerator.this, j, o);
                    this.codeWrapArgument(this.parameterTypes[j], array[j], o);
                    o.writeByte(83);
                }
            }
            else {
                o.writeByte(1);
            }
            o.writeByte(185);
            o.writeShort((int)ProxyGenerator.this.cp.getInterfaceMethodRef("java/lang/reflect/InvocationHandler", "invoke", "(Ljava/lang/Object;Ljava/lang/reflect/Method;[Ljava/lang/Object;)Ljava/lang/Object;"));
            o.writeByte(4);
            o.writeByte(0);
            if (this.returnType == Void.TYPE) {
                o.writeByte(87);
                o.writeByte(177);
            }
            else {
                this.codeUnwrapReturnValue(this.returnType, o);
            }
            int n4;
            int n3 = n4 = (short)methodInfo.code.size();
            List<Object> access$1100 = ProxyGenerator.access$1100(this.exceptionTypes);
            if (access$1100.size() > 0) {
                Iterator<Object> iterator = access$1100.iterator();
                while (iterator.hasNext()) {
                    methodInfo.exceptionTable.add(new ProxyGenerator.ExceptionTableEntry((short)n2, (short)n4, (short)n3, ProxyGenerator.this.cp.getClass(ProxyGenerator.access$900(((Class)iterator.next()).getName()))));
                }
                o.writeByte(191);
                methodInfo.exceptionTable.add(new ProxyGenerator.ExceptionTableEntry((short)n2, (short)n4, (short)methodInfo.code.size(), ProxyGenerator.this.cp.getClass("java/lang/Throwable")));
                ProxyGenerator.access$1200(ProxyGenerator.this, n, o);
                o.writeByte(187);
                o.writeShort((int)ProxyGenerator.this.cp.getClass("java/lang/reflect/UndeclaredThrowableException"));
                o.writeByte(89);
                ProxyGenerator.access$700(ProxyGenerator.this, n, o);
                o.writeByte(183);
                o.writeShort((int)ProxyGenerator.this.cp.getMethodRef("java/lang/reflect/UndeclaredThrowableException", "<init>", "(Ljava/lang/Throwable;)V"));
                o.writeByte(191);
            }
            if (methodInfo.code.size() > 65535) {
                throw new IllegalArgumentException("code size limit exceeded");
            }
            methodInfo.maxStack = 10;
            methodInfo.maxLocals = (short)(n + 1);
            methodInfo.declaredExceptions = new short[this.exceptionTypes.length];
            for (int k = 0; k < this.exceptionTypes.length; ++k) {
                methodInfo.declaredExceptions[k] = ProxyGenerator.this.cp.getClass(ProxyGenerator.access$900(this.exceptionTypes[k].getName()));
            }
            return methodInfo;
        }
        
        private void codeWrapArgument(Class<?> clazz, int n, DataOutputStream dataOutputStream) throws IOException {
            if (clazz.isPrimitive()) {
                ProxyGenerator.PrimitiveTypeInfo value = ProxyGenerator.PrimitiveTypeInfo.get(clazz);
                if (clazz == Integer.TYPE || clazz == Boolean.TYPE || clazz == Byte.TYPE || clazz == Character.TYPE || clazz == Short.TYPE) {
                    ProxyGenerator.access$1300(ProxyGenerator.this, n, dataOutputStream);
                }
                else if (clazz == Long.TYPE) {
                    ProxyGenerator.access$1400(ProxyGenerator.this, n, dataOutputStream);
                }
                else if (clazz == Float.TYPE) {
                    ProxyGenerator.access$1500(ProxyGenerator.this, n, dataOutputStream);
                }
                else if (clazz == Double.TYPE) {
                    ProxyGenerator.access$1600(ProxyGenerator.this, n, dataOutputStream);
                }
                else {
                    throw new AssertionError();
                }
                dataOutputStream.writeByte(184);
                dataOutputStream.writeShort((int)ProxyGenerator.this.cp.getMethodRef(value.wrapperClassName, "valueOf", value.wrapperValueOfDesc));
            }
            else {
                ProxyGenerator.access$700(ProxyGenerator.this, n, dataOutputStream);
            }
        }
        
        private void codeUnwrapReturnValue(Class<?> clazz, DataOutputStream dataOutputStream) throws IOException {
            if (clazz.isPrimitive()) {
                ProxyGenerator.PrimitiveTypeInfo value = ProxyGenerator.PrimitiveTypeInfo.get(clazz);
                dataOutputStream.writeByte(192);
                dataOutputStream.writeShort((int)ProxyGenerator.this.cp.getClass(value.wrapperClassName));
                dataOutputStream.writeByte(182);
                dataOutputStream.writeShort((int)ProxyGenerator.this.cp.getMethodRef(value.wrapperClassName, value.unwrapMethodName, value.unwrapMethodDesc));
                if (clazz == Integer.TYPE || clazz == Boolean.TYPE || clazz == Byte.TYPE || clazz == Character.TYPE || clazz == Short.TYPE) {
                    dataOutputStream.writeByte(172);
                }
                else if (clazz == Long.TYPE) {
                    dataOutputStream.writeByte(173);
                }
                else if (clazz == Float.TYPE) {
                    dataOutputStream.writeByte(174);
                }
                else if (clazz == Double.TYPE) {
                    dataOutputStream.writeByte(175);
                }
                else {
                    throw new AssertionError();
                }
            }
            else {
                dataOutputStream.writeByte(192);
                dataOutputStream.writeShort((int)ProxyGenerator.this.cp.getClass(ProxyGenerator.access$900(clazz.getName())));
                dataOutputStream.writeByte(176);
            }
        }
        
        private void codeFieldInitialization(DataOutputStream dataOutputStream) throws IOException {
            ProxyGenerator.access$1700(ProxyGenerator.this, this.fromClass, dataOutputStream);
            ProxyGenerator.access$1800(ProxyGenerator.this, (int)ProxyGenerator.this.cp.getString(this.methodName), dataOutputStream);
            ProxyGenerator.access$1000(ProxyGenerator.this, this.parameterTypes.length, dataOutputStream);
            dataOutputStream.writeByte(189);
            dataOutputStream.writeShort((int)ProxyGenerator.this.cp.getClass("java/lang/Class"));
            for (int i = 0; i < this.parameterTypes.length; ++i) {
                dataOutputStream.writeByte(89);
                ProxyGenerator.access$1000(ProxyGenerator.this, i, dataOutputStream);
                if (this.parameterTypes[i].isPrimitive()) {
                    ProxyGenerator.PrimitiveTypeInfo value = ProxyGenerator.PrimitiveTypeInfo.get(this.parameterTypes[i]);
                    dataOutputStream.writeByte(178);
                    dataOutputStream.writeShort((int)ProxyGenerator.this.cp.getFieldRef(value.wrapperClassName, "TYPE", "Ljava/lang/Class;"));
                }
                else {
                    ProxyGenerator.access$1700(ProxyGenerator.this, this.parameterTypes[i], dataOutputStream);
                }
                dataOutputStream.writeByte(83);
            }
            dataOutputStream.writeByte(182);
            dataOutputStream.writeShort((int)ProxyGenerator.this.cp.getMethodRef("java/lang/Class", "getMethod", "(Ljava/lang/String;[Ljava/lang/Class;)Ljava/lang/reflect/Method;"));
            dataOutputStream.writeByte(179);
            dataOutputStream.writeShort((int)ProxyGenerator.this.cp.getFieldRef(ProxyGenerator.access$900(ProxyGenerator.this.className), this.methodFieldName, "Ljava/lang/reflect/Method;"));
        }
        
        static ProxyGenerator.MethodInfo access$100(ProxyGenerator.ProxyMethod proxyMethod) throws IOException {
            return proxyMethod.generateMethod();
        }
        
        static void access$1900(ProxyGenerator.ProxyMethod proxyMethod, DataOutputStream dataOutputStream) throws IOException {
            proxyMethod.codeFieldInitialization(dataOutputStream);
        }
    }
    
    private static class PrimitiveTypeInfo
    {
        public String baseTypeString;
        public String wrapperClassName;
        public String wrapperValueOfDesc;
        public String unwrapMethodName;
        public String unwrapMethodDesc;
        private static Map<Class<?>, ProxyGenerator.PrimitiveTypeInfo> table;
        
        private static void add(Class<?> clazz, Class<?> clazz2) {
            ProxyGenerator.PrimitiveTypeInfo.table.put(clazz, new PrimitiveTypeInfo(clazz, clazz2));
        }
        
        private PrimitiveTypeInfo(Class<?> componentType, Class<?> clazz) {
            assert !(!componentType.isPrimitive());
            this.baseTypeString = Array.newInstance(componentType, 0).getClass().getName().substring(1);
            this.wrapperClassName = ProxyGenerator.access$900(clazz.getName());
            this.wrapperValueOfDesc = "(" + this.baseTypeString + ")L" + this.wrapperClassName + ";";
            this.unwrapMethodName = componentType.getName() + "Value";
            this.unwrapMethodDesc = "()" + this.baseTypeString;
            return;
        }
        
        public static ProxyGenerator.PrimitiveTypeInfo get(Class<?> clazz) {
            return (ProxyGenerator.PrimitiveTypeInfo)ProxyGenerator.PrimitiveTypeInfo.table.get(clazz);
        }
        
        static {
            ProxyGenerator.PrimitiveTypeInfo.table = new HashMap();
            add(Byte.TYPE, Byte.class);
            add(Character.TYPE, Character.class);
            add(Double.TYPE, Double.class);
            add(Float.TYPE, Float.class);
            add(Integer.TYPE, Integer.class);
            add(Long.TYPE, Long.class);
            add(Short.TYPE, Short.class);
            add(Boolean.TYPE, Boolean.class);
        }
    }
    
    private static class ConstantPool
    {
        private List<ProxyGenerator.ConstantPool.Entry> pool;
        private Map<Object, Short> map;
        private boolean readOnly;
        
        private ConstantPool() {
            this.pool = new ArrayList(32);
            this.map = new HashMap(16);
            this.readOnly = false;
        }
        
        public short getUtf8(String s) {
            if (s == null) {
                throw new NullPointerException();
            }
            return this.getValue(s);
        }
        
        public short getInteger(int value) {
            return this.getValue(new Integer(value));
        }
        
        public short getFloat(float value) {
            return this.getValue(new Float(value));
        }
        
        public short getClass(String s) {
            return this.getIndirect(new ProxyGenerator.ConstantPool.IndirectEntry(7, this.getUtf8(s)));
        }
        
        public short getString(String s) {
            return this.getIndirect(new ProxyGenerator.ConstantPool.IndirectEntry(8, this.getUtf8(s)));
        }
        
        public short getFieldRef(String s, String s2, String s3) {
            return this.getIndirect(new ProxyGenerator.ConstantPool.IndirectEntry(9, this.getClass(s), this.getNameAndType(s2, s3)));
        }
        
        public short getMethodRef(String s, String s2, String s3) {
            return this.getIndirect(new ProxyGenerator.ConstantPool.IndirectEntry(10, this.getClass(s), this.getNameAndType(s2, s3)));
        }
        
        public short getInterfaceMethodRef(String s, String s2, String s3) {
            return this.getIndirect(new ProxyGenerator.ConstantPool.IndirectEntry(11, this.getClass(s), this.getNameAndType(s2, s3)));
        }
        
        public short getNameAndType(String s, String s2) {
            return this.getIndirect(new ProxyGenerator.ConstantPool.IndirectEntry(12, this.getUtf8(s), this.getUtf8(s2)));
        }
        
        public void setReadOnly() {
            this.readOnly = true;
        }
        
        public void write(OutputStream out) throws IOException {
            DataOutputStream dataOutputStream = new DataOutputStream(out);
            dataOutputStream.writeShort(this.pool.size() + 1);
            Iterator<ProxyGenerator.ConstantPool.Entry> iterator = this.pool.iterator();
            while (iterator.hasNext()) {
                ((ProxyGenerator.ConstantPool.Entry)iterator.next()).write(dataOutputStream);
            }
        }
        
        private short addEntry(ProxyGenerator.ConstantPool.Entry entry) {
            this.pool.add(entry);
            if (this.pool.size() >= 65535) {
                throw new IllegalArgumentException("constant pool size limit exceeded");
            }
            return (short)this.pool.size();
        }
        
        private short getValue(Object obj) {
            Short n = (Short)this.map.get(obj);
            if (n != null) {
                return (short)n;
            }
            if (this.readOnly) {
                throw new InternalError("late constant pool addition: " + obj);
            }
            int addEntry = this.addEntry(new ValueEntry(obj));
            this.map.put(obj, new Short((short)addEntry));
            return addEntry;
        }
        
        private short getIndirect(ProxyGenerator.ConstantPool.IndirectEntry indirectEntry) {
            Short n = (Short)this.map.get(indirectEntry);
            if (n != null) {
                return (short)n;
            }
            if (this.readOnly) {
                throw new InternalError("late constant pool addition");
            }
            int addEntry = this.addEntry(indirectEntry);
            this.map.put(indirectEntry, new Short((short)addEntry));
            return addEntry;
        }
        
        private abstract static class Entry
        {
            public abstract void write(DataOutputStream p0) throws IOException;
        }
        
        private static class ValueEntry extends ProxyGenerator.ConstantPool.Entry
        {
            private Object value;
            
            public ValueEntry(Object value) {
                this.value = value;
            }
            
            @Override
            public void write(DataOutputStream dataOutputStream) throws IOException {
                if ((this.value instanceof String)) {
                    dataOutputStream.writeByte(1);
                    dataOutputStream.writeUTF((String)this.value);
                }
                else if ((this.value instanceof Integer)) {
                    dataOutputStream.writeByte(3);
                    dataOutputStream.writeInt((int)this.value);
                }
                else if ((this.value instanceof Float)) {
                    dataOutputStream.writeByte(4);
                    dataOutputStream.writeFloat((float)this.value);
                }
                else if ((this.value instanceof Long)) {
                    dataOutputStream.writeByte(5);
                    dataOutputStream.writeLong((long)this.value);
                }
                else if ((this.value instanceof Double)) {
                    dataOutputStream.writeDouble(6.0);
                    dataOutputStream.writeDouble((double)this.value);
                }
                else {
                    throw new InternalError("bogus value entry: " + this.value);
                }
            }
        }
        
        private static class IndirectEntry extends ProxyGenerator.ConstantPool.Entry
        {
            private int tag;
            private short index0;
            private short index1;
            
            public IndirectEntry(int tag, short index0) {
                this.tag = tag;
                this.index0 = index0;
                this.index1 = 0;
            }
            
            public IndirectEntry(int tag, short index0, short index2) {
                this.tag = tag;
                this.index0 = index0;
                this.index1 = index2;
            }
            
            @Override
            public void write(DataOutputStream dataOutputStream) throws IOException {
                dataOutputStream.writeByte(this.tag);
                dataOutputStream.writeShort((int)this.index0);
                if (this.tag == 9 || this.tag == 10 || this.tag == 11 || this.tag == 12) {
                    dataOutputStream.writeShort((int)this.index1);
                }
            }
            
            @Override
            public int hashCode() {
                return this.tag + this.index0 + this.index1;
            }
            
            @Override
            public boolean equals(Object o) {
                if ((o instanceof ProxyGenerator.ConstantPool.IndirectEntry)) {
                    ProxyGenerator.ConstantPool.IndirectEntry indirectEntry = (ProxyGenerator.ConstantPool.IndirectEntry)o;
                    if (this.tag == indirectEntry.tag && this.index0 == indirectEntry.index0 && this.index1 == indirectEntry.index1) {
                        return true;
                    }
                }
                return false;
            }
        }
    }
}
